/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2016 Red Hat, Inc.
 * All rights reserved.
 *
 * License: GPL (version 3 or any later version).
 * See LICENSE for details. 
 * END COPYRIGHT BLOCK **/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

/*
 * AD DN bind plugin
 *
 * This plugin allows violation of the LDAP v3 standard for binds to occur
 * as "uid" or "uid@example.com". This plugin re-maps those terms to a real
 * dn such as uid=user,ou=People,dc=example,dc=com.
 */

#include "addn.h"

/* is there a better type we can use here? */
#define ADDN_FAILURE 1
#define SLAPI_LOG_PLUGIN 0

static void* plugin_identity = NULL;

static Slapi_PluginDesc addn_description = {
    "addn",
    VENDOR,
    DS_PACKAGE_VERSION,
    "Allow AD DN style bind names to LDAP"
};

int
addn_filter_validate(char *config_filter)
{
    /* Copy the filter */
    size_t slen = strlen(config_filter);
    int found = 0;
    int percent_last = 0;

    for (size_t i = 0; i < slen; i++) {
        if (percent_last == 1 && config_filter[i] == 's') {
            found += 1;
            percent_last = 0;
        } else if (percent_last == 0 && config_filter[i] == '%') {
            percent_last = 1;
        } else {
            percent_last = 0;
        }
    }

    if (found == 1) {
        return LDAP_SUCCESS;
    }

    /* Is there a way to now compile and check the filter syntax? */

    return 1;
}

int
addn_prebind(Slapi_PBlock *pb)
{
    // Slapi_Entry *plugin_entry = NULL;
    struct addn_config *config = NULL;
    Slapi_DN *pb_sdn_target = NULL;
    Slapi_DN *pb_sdn_mapped = NULL;
    const char *dn = NULL;

    const char *dn_bind = NULL;
    int dn_bind_len = 0;
    char *dn_bind_escaped = NULL;

    const char *dn_domain = NULL;
    int dn_domain_len = 0;
    char *dn_domain_escaped = NULL;

    char *be_suffix = NULL;
    Slapi_DN *be_suffix_dn = NULL;
    char *config_filter = NULL;
    char *filter = NULL;
    int result = 0;

    Slapi_PBlock *search_pblock = NULL;
    int search_result = 0;
    Slapi_Entry **entries = NULL;
    int entry_count = 0;

    slapi_log_error(SLAPI_LOG_PLUGIN, "addn_prebind", "--> begin\n");
    /* This is taking .2 of a second on my laptop .... */
    /* Perhaps it's because I'm using -O0? */
    slapi_pblock_get(pb, SLAPI_BIND_TARGET_SDN, &pb_sdn_target);
    dn = slapi_sdn_get_dn(pb_sdn_target);
    if (dn == NULL) {
        result = ADDN_FAILURE;
        goto out;
    }
    slapi_log_error(SLAPI_LOG_PLUGIN, "addn_prebind", "Recieved %s\n", dn);

    result = slapi_dn_syntax_check(NULL, dn, 0);
    slapi_log_error(SLAPI_LOG_PLUGIN, "addn_prebind", "Dn validation %d\n", result);

    if (result == LDAP_SUCCESS) {
        slapi_log_error(SLAPI_LOG_PLUGIN, "addn_prebind", "Dn syntax is correct, do not alter\n");
        /* This is a syntax correct bind dn. Leave it alone! */
        goto out;
    }
    slapi_log_error(SLAPI_LOG_PLUGIN, "addn_prebind", "Dn syntax is incorrect, it may need ADDN mangaling\n");


    result = slapi_pblock_get(pb, SLAPI_PLUGIN_PRIVATE, &config);
    if (result != LDAP_SUCCESS || config == NULL) {
        slapi_log_error(SLAPI_LOG_FATAL, "addn_prebind", "CRITICAL: Unable to retrieve plugin configuration!\n");
        result = ADDN_FAILURE;
        goto out;
    }

    /* Right, the dn is invalid! This means it *could* be addn syntax */
    /* Do we have any other validation to run here? */
    /* Split the @domain component (if any) */

    dn_bind = strtok(dn, "@");

    /* Filter escape the name, and the domain parts */
    if (dn_bind != NULL) {
        dn_bind_len = strlen(dn_bind);
        dn_bind_escaped = slapi_escape_filter_value(dn_bind, dn_bind_len);
        dn_bind_len = strlen(dn_bind_escaped);
    }

    dn_domain = strtok(NULL, "@");

    /* The handling of user@domain isn't done yet: This code is a placeholder */
    if (dn_domain != NULL) {
        dn_domain_len = strlen(dn_domain);
        dn_domain_escaped = slapi_escape_filter_value(dn_domain, dn_domain_len);
        dn_domain_len = strlen(dn_domain_escaped);
    }

    /* If we don't have the @domain. get a default from config */

    if (dn_domain_escaped == NULL) {
        /* This could alternately be domain, then we do the same domain -> suffix conversion .... */
        if (config->default_suffix == NULL) {
            slapi_log_error(SLAPI_LOG_FATAL, "addn_prebind", "CRITICAL: No default backend in configuration, you must set addn_default_suffix\n");
            result = ADDN_FAILURE;
            goto out;
        }
        /* Select the correct backend. If no backend, bail. */
        slapi_log_error(SLAPI_LOG_PLUGIN, "addn_prebind", "Selected suffix %s\n", config->default_suffix);
        be_suffix_dn = slapi_sdn_new_dn_byval(config->default_suffix);
    } else {
        /* Future: Parse the domain out to a potential suffix name. */
        slapi_log_error(SLAPI_LOG_PLUGIN, "addn_prebind", "dn_domain_escaped %s\n", dn_domain_escaped);
        /* I think for now, we assume that a.b.c => dc=a,dc=b,dc=c */
        size_t be_suffix_len = 1;
        size_t offset = 0;
        be_suffix = slapi_ch_malloc(sizeof(char) * be_suffix_len);
        char *token = strtok(dn_domain_escaped, ".");
        while (token != NULL) {
            /* check len of the token. */
            size_t len = strlen(token);
            /* Expand be_suffix by len + 4 (,dc=) */
            be_suffix_len = be_suffix_len + len + 4;
            be_suffix = slapi_ch_realloc(be_suffix, be_suffix_len);
            /* Inject our content at the start. */
            strncpy(be_suffix + offset, ",dc=", 4);
            strncpy(be_suffix + offset + 4, token, len);
            /* bump the offset */
            offset = offset + len + 4;
            be_suffix[offset] = '\0';
            token = strtok(NULL, ".");
        }
        /* Force the last char to be null. */
        slapi_log_error(SLAPI_LOG_PLUGIN, "addn_prebind", "Domain %s to suffix %s\n", dn_domain_escaped, be_suffix + 1);
        /* Start at +1 to trim the leading , off. */
        be_suffix_dn = slapi_sdn_new_dn_byval(be_suffix + 1);
        slapi_ch_free_string(&be_suffix);
    }

    /* Get our filter. From the config */
    config_filter = config->filter_template;

    /* Put the search term into the filter. */
    //!! How does this return errors, like not enough values? Too many?
    //!! This needs a careful check and review!!!
    filter = slapi_ch_smprintf(config_filter, dn_bind_escaped);
    slapi_log_error(SLAPI_LOG_PLUGIN, "addn_prebind", "Searching with filter %s\n", filter);

    /* Search the backend, and apply our filter given the username */
    search_pblock = slapi_pblock_new();

    if (search_pblock == NULL) {
        slapi_log_error(SLAPI_LOG_FATAL, "addn_prebind", "CRITICAL: Unable to allocate search_pblock!!!\n");
        result = ADDN_FAILURE;
        goto out;
    }

    // How can I do this search to retrieve NO attributes, only the DN?

    slapi_search_internal_set_pb_ext(search_pblock, be_suffix_dn, LDAP_SCOPE_SUBTREE,
                                     filter, NULL, 0 /* attrs only */,
                                     NULL /* controls */, NULL /* uniqueid */, 
                                     plugin_identity, 0 /* actions */);
    slapi_search_internal_pb(search_pblock);

    result = slapi_pblock_get(search_pblock, SLAPI_PLUGIN_INTOP_RESULT, &search_result);
    if (result != LDAP_SUCCESS) {
        slapi_log_error(SLAPI_LOG_FATAL, "addn_prebind", "CRITICAL: Internal search pblock get failed!!!\n");
        result = ADDN_FAILURE;
        goto out;
    }

    /* No results are an error: Allow the plugin to continue, because */
    /* the bind will fail anyway ....? */

    /* Do we want to return a fail here? Or pass, and allow the auth code
     * to fail us. That will prevent some attacks I feel as this could be a
     * disclosure attack ....
     */
    if (search_result == LDAP_NO_SUCH_OBJECT) {
        slapi_log_error(SLAPI_LOG_PLUGIN, "addn_prebind", "LDAP_NO_SUCH_OBJECT \n");
        result = LDAP_SUCCESS;
        goto out;
    }

    /* On all other errors, just fail out. */
    if (search_result != LDAP_SUCCESS) {
        slapi_log_error(SLAPI_LOG_PLUGIN, "addn_prebind", "CRITICAL: Internal search error occured %d \n", search_result);
        result = ADDN_FAILURE;
        goto out;
    }


    result = slapi_pblock_get(search_pblock, SLAPI_NENTRIES, &entry_count);

    if (result != LDAP_SUCCESS) {
        slapi_log_error(SLAPI_LOG_FATAL, "addn_prebind", "CRITICAL: Unable to retrieve number of entries from pblock!\n");
        result = ADDN_FAILURE;
        goto out;
    }
    /* If there are multiple results, we should also fail as we cannot */
    /* be sure of the real object we want to bind to */

    if (entry_count > 1) {
        /*  Is there a better log level for this? */
        slapi_log_error(SLAPI_LOG_FATAL, "addn_prebind", "WARNING, multiple results returned. Failing to auth ...\n");
        result = ADDN_FAILURE;
        goto out;
    }

    result = slapi_pblock_get(search_pblock, SLAPI_PLUGIN_INTOP_SEARCH_ENTRIES, &entries);
    if (result != LDAP_SUCCESS) {
        slapi_log_error(SLAPI_LOG_FATAL, "addn_prebind", "CRITICAL: Unable to retrieve entries from pblock!\n");
        result = ADDN_FAILURE;
        goto out;
    }

    /* We should only have one entry here! */

    pb_sdn_mapped = slapi_sdn_dup(slapi_entry_get_sdn(*entries));

    slapi_log_error(SLAPI_LOG_PLUGIN, "addn_prebind", "SEARCH entry dn=%s is mapped from addn=%s\n", slapi_sdn_get_dn(pb_sdn_mapped), dn);


    /* If there is a result, take the DN and make the SDN for PB */

    /* Free the original SDN */
    result = slapi_pblock_set(pb, SLAPI_TARGET_SDN, pb_sdn_mapped);
    if (result != LDAP_SUCCESS) {
        slapi_log_error(SLAPI_LOG_FATAL, "addn_prebind", "CRITICAL: Unable to set new mapped DN to pblock!\n");
        /* We have to free the mapped SDN here */
        slapi_sdn_free(&pb_sdn_mapped);
        result = ADDN_FAILURE;
        goto out;
    }

    slapi_sdn_free(&pb_sdn_target);

    slapi_log_error(SLAPI_LOG_PLUGIN, "addn_prebind", "<-- complete\n");

    result = LDAP_SUCCESS;

out:
    if (search_pblock != NULL) {
        slapi_free_search_results_internal(search_pblock);
        slapi_pblock_destroy(search_pblock);
    }
    slapi_sdn_free(&be_suffix_dn);
    slapi_ch_free_string(&dn_bind_escaped);
    slapi_ch_free_string(&dn_domain_escaped);
    slapi_ch_free_string(&filter);

    return result;
}


/*
 * addn_start
 *
 * This is called when the plugin is started. It allows a configuration change
 * To be made while the directory server is live.
 */
int
addn_start(Slapi_PBlock *pb)
{
    Slapi_Entry *plugin_entry = NULL;
    struct addn_config *config = NULL;
    int result = LDAP_SUCCESS;

    slapi_log_error(SLAPI_LOG_PLUGIN, "addn_start", "starting ...\n");

    /* I have no idea how this works .... should be SLAPI_PLUGIN_CONFIG */
    result = slapi_pblock_get(pb, SLAPI_ADD_ENTRY, &plugin_entry);
    if (result != LDAP_SUCCESS || plugin_entry == NULL) {
        slapi_log_error(SLAPI_LOG_FATAL, "addn_start", "CRITICAL: Failed to retrieve config entry!\n");
        return SLAPI_PLUGIN_FAILURE;
    }

    /* Now set the values into the config */
    /* Are there some error cases here we should be checking? */
    config = (struct addn_config *)slapi_ch_malloc(sizeof(struct addn_config));
    config->default_suffix = slapi_entry_attr_get_charptr(plugin_entry, "addn_default_suffix");
    config->filter_template = slapi_entry_attr_get_charptr(plugin_entry, "addn_filter");

    /* Validate the filter_template only has one %s */
    if (addn_filter_validate(config->filter_template) != LDAP_SUCCESS) {
        slapi_log_error(SLAPI_LOG_FATAL, "addn_start", "CRITICAL: Failed to validate addn_filter config entry!\n");
        slapi_ch_free_string(&config->default_suffix);
        slapi_ch_free_string(&config->filter_template);
        slapi_ch_free((void **) &config);
        return SLAPI_PLUGIN_FAILURE;
    }

    /* Set into the pblock */
    slapi_pblock_set(pb, SLAPI_PLUGIN_PRIVATE, (void *) config);

    slapi_log_error(SLAPI_LOG_PLUGIN, "addn_start", "startup complete\n");

    return LDAP_SUCCESS;
}

/*
 * addn_close
 *
 * Called when the plugin is stopped. Frees the configs as needed
 */
int
addn_close(Slapi_PBlock *pb)
{
    struct addn_config *config = NULL;

    slapi_log_error(SLAPI_LOG_PLUGIN, "addn_close", "stopping ...\n");

    slapi_pblock_get(pb, SLAPI_PLUGIN_PRIVATE, &config);
    if (config != NULL) {
        slapi_ch_free_string(&config->default_suffix);
        slapi_ch_free_string(&config->filter_template);
        slapi_ch_free((void **) &config);
        slapi_pblock_set(pb, SLAPI_PLUGIN_PRIVATE, NULL);
    }

    slapi_log_error(SLAPI_LOG_PLUGIN, "addn_close", "stop complete\n");

    return LDAP_SUCCESS;
}


/*
 * addn_init
 *
 * Initialise and register the addn plugin to the directory server.
 */
int
addn_init(Slapi_PBlock *pb)
{
    int result = 0;

    result = slapi_pblock_set(pb, SLAPI_PLUGIN_VERSION, SLAPI_PLUGIN_VERSION_03);
    if (result != LDAP_SUCCESS) {
        goto out;
    }

    /* Get and stash our plugin identity */
    slapi_pblock_get (pb, SLAPI_PLUGIN_IDENTITY, &plugin_identity);

    result = slapi_pblock_set(pb, SLAPI_PLUGIN_DESCRIPTION, (void*)&addn_description);
    if (result != LDAP_SUCCESS) {
        goto out;
    }

    result = slapi_pblock_set(pb, SLAPI_PLUGIN_START_FN, (void*)addn_start);
    if (result != LDAP_SUCCESS) {
        goto out;
    }

    result = slapi_pblock_set(pb, SLAPI_PLUGIN_CLOSE_FN, (void*)addn_close);
    if (result != LDAP_SUCCESS) {
        goto out;
    }

    result = slapi_pblock_set(pb, SLAPI_PLUGIN_PRE_BIND_FN, (void*)addn_prebind);
    if (result != LDAP_SUCCESS) {
        goto out;
    }

out:
    if (result == LDAP_SUCCESS) {
        slapi_log_error(SLAPI_LOG_PLUGIN, "addn_init", "Success: plugin loaded.\n");
        slapi_log_error(SLAPI_LOG_FATAL, "addn_init", "Warning: The use of this plugin violates the LDAPv3 specification RFC4511 section 4.2 BindDN specification. You have been warned ...\n");
    } else {
        slapi_log_error(SLAPI_LOG_PLUGIN, "addn_init", "Error: %d. \n", result);
    }
    return result;
}


